#include <linux/blkdev.h>
#include <linux/delay.h>
#include <linux/idr.h>
#include <linux/init.h>
#include <linux/module.h>
#include <linux/kernel.h>
#include <linux/pci.h>
#include <linux/nvme.h>
#include <linux/io-64-nonatomic-lo-hi.h>

MODULE_LICENSE("GPL");

struct nvme_cmb_dev {
	void __iomem *bar;
	void __iomem *buffer_bar;
	size_t buffer_offset;
	size_t buffer_size;

	struct gendisk *disk;
	int index;

	struct nvme_command *admin_sqes;
	dma_addr_t admin_sq_dma_addr;
	struct nvme_completion *admin_cqes;
	dma_addr_t admin_cq_dma_addr;

	u32 cc;
};

#define QUEUE_SIZE 2

static DEFINE_IDA(nvme_cmb_index_ida);

static void nvme_cmb_submit_bio(struct bio *bio)
{
	struct nvme_cmb_dev *dev = bio->bi_bdev->bd_disk->private_data;
	void __iomem *buffer = dev->buffer_bar + dev->buffer_offset;
	struct bvec_iter iter;
	struct bio_vec bv;

	bio_for_each_segment(bv, bio, iter) {
		void __iomem *dev_addr = buffer +
		    (iter.bi_sector << SECTOR_SHIFT);
		void *kaddr = bvec_virt(&bv);

		if (bio_op(bio) == REQ_OP_READ)
			memcpy_fromio(kaddr, dev_addr, bv.bv_len);
		else
			memcpy_toio(dev_addr, kaddr, bv.bv_len);
	}

	bio_endio(bio);
}

static const struct block_device_operations nvme_cmb_ops = {
	.owner = THIS_MODULE,
	.submit_bio = nvme_cmb_submit_bio,
};

static int nvme_cmb_alloc_queues(struct nvme_cmb_dev *dev, struct pci_dev *pdev)
{
	dev->admin_sqes = dma_alloc_coherent(&pdev->dev,
					     QUEUE_SIZE *
					     sizeof(*dev->admin_sqes),
					     &dev->admin_sq_dma_addr,
					     GFP_KERNEL);
	if (!dev->admin_sqes)
		return -ENOMEM;
	lo_hi_writeq(dev->admin_sq_dma_addr, dev->bar + NVME_REG_ASQ);

	dev->admin_cqes = dma_alloc_coherent(&pdev->dev,
					     QUEUE_SIZE *
					     sizeof(*dev->admin_cqes),
					     &dev->admin_cq_dma_addr,
					     GFP_KERNEL);
	if (!dev->admin_cqes) {
		dma_free_coherent(&pdev->dev,
				  QUEUE_SIZE * sizeof(*dev->admin_cqes),
				  dev->admin_cqes, dev->admin_cq_dma_addr);
		return -ENOMEM;
	}
	lo_hi_writeq(dev->admin_cq_dma_addr, dev->bar + NVME_REG_ACQ);

	writel((QUEUE_SIZE - 1) | ((QUEUE_SIZE - 1) << 16),
	       dev->bar + NVME_REG_AQA);
	return 0;
}

static void nvme_cmb_free_queues(struct nvme_cmb_dev *dev, struct pci_dev *pdev)
{
	dma_free_coherent(&pdev->dev, QUEUE_SIZE * sizeof(*dev->admin_cqes),
			  dev->admin_cqes, dev->admin_cq_dma_addr);
	dma_free_coherent(&pdev->dev, QUEUE_SIZE * sizeof(*dev->admin_sqes),
			  dev->admin_sqes, dev->admin_sq_dma_addr);
}

static void nvme_cmb_enable(struct nvme_cmb_dev *dev)
{
	dev->cc = NVME_CC_CSS_NVM | NVME_CC_AMS_RR |
	    NVME_CC_IOSQES | NVME_CC_IOCQES;
	writel(dev->cc, dev->bar + NVME_REG_CC);

	dev->cc |= NVME_CC_ENABLE;
	writel(dev->cc, dev->bar + NVME_REG_CC);
	while (true) {
		u32 csts = readl(dev->bar + NVME_REG_CSTS);

		if (csts & NVME_CSTS_RDY)
			break;
		usleep_range(1000, 2000);
	}
}

static void nvme_cmb_shutdown(struct nvme_cmb_dev *dev)
{
	dev->cc &= ~NVME_CC_SHN_MASK;
	dev->cc |= NVME_CC_SHN_NORMAL;
	writel(dev->cc, dev->bar + NVME_REG_CC);
	while (true) {
		u32 csts = readl(dev->bar + NVME_REG_CSTS);

		if ((csts & NVME_CSTS_SHST_MASK) == NVME_CSTS_SHST_CMPLT)
			break;
		usleep_range(1000, 2000);
	}
}

static int nvme_cmb_map_buffer(struct nvme_cmb_dev *dev, struct pci_dev *pdev)
{
	u32 cmbloc = readl(dev->bar + NVME_REG_CMBLOC);
	u32 cmbsz = readl(dev->bar + NVME_REG_CMBSZ);
	u64 unit = 1ULL << (12 + 4 *
			    ((cmbsz >> NVME_CMBSZ_SZU_SHIFT) &
			     NVME_CMBSZ_SZU_MASK));

	dev->buffer_bar = pci_ioremap_bar(pdev, NVME_CMB_BIR(cmbloc));
	if (!dev->buffer_bar)
		return -ENOMEM;

	dev->buffer_offset = unit * NVME_CMB_OFST(cmbloc);
	dev->buffer_size =
	    unit * ((cmbsz >> NVME_CMBSZ_SZ_SHIFT) & NVME_CMBSZ_SZ_MASK);
	return 0;
}

static int nvme_cmb_probe(struct pci_dev *pdev, const struct pci_device_id *id)
{
	struct nvme_cmb_dev *dev;
	int error;

	dev_info(&pdev->dev, "found NVMe device\n");

	dev = kzalloc(sizeof(*dev), GFP_KERNEL);
	if (!dev)
		return -ENOMEM;

	pci_set_master(pdev);
	error = pci_enable_device_mem(pdev);
	if (error)
		goto out_free_dev;
	error = pci_request_mem_regions(pdev, "nvme-cmb");
	if (error)
		goto out_disable_device;

	dev->bar = pci_ioremap_bar(pdev, 0);
	if (!dev->bar) {
		error = -ENOMEM;
		goto out_release_regions;
	}

	if (readl(dev->bar + NVME_REG_CSTS) == ~0) {
		error = -EIO;
		goto out_iounmap;
	}

	error = nvme_cmb_alloc_queues(dev, pdev);
	if (error)
		goto out_iounmap;

	nvme_cmb_enable(dev);

	error = nvme_cmb_map_buffer(dev, pdev);
	if (error)
		goto out_shutdown;

	dev->disk = blk_alloc_disk(NUMA_NO_NODE);
	if (!dev->disk) {
		error = -ENOMEM;
		goto out_unmap_cmb;
	}
	dev->disk->private_data = dev;
	dev->disk->fops = &nvme_cmb_ops;
	set_capacity(dev->disk, dev->buffer_size >> SECTOR_SHIFT);

	error = ida_alloc(&nvme_cmb_index_ida, GFP_KERNEL);
	if (error < 0)
		goto out_put_disk;
	snprintf(dev->disk->disk_name, DISK_NAME_LEN, "nc%d", error);

	error = device_add_disk(&pdev->dev, dev->disk, NULL);
	if (error)
		goto out_ida_free;

	dev_set_drvdata(&pdev->dev, dev);
	dev_info(&pdev->dev, "added %zu MiB ramdisk\n",
		 dev->buffer_size / SZ_1M);
	return 0;

out_ida_free:
	ida_free(&nvme_cmb_index_ida, dev->index);
out_put_disk:
	put_disk(dev->disk);
out_unmap_cmb:
	iounmap(dev->buffer_bar);
out_shutdown:
	nvme_cmb_shutdown(dev);
	nvme_cmb_free_queues(dev, pdev);
out_iounmap:
	iounmap(dev->bar);
out_release_regions:
	pci_release_mem_regions(pdev);
out_disable_device:
	pci_disable_device(pdev);
out_free_dev:
	kfree(dev);
	return error;
}

static void nvme_cmb_remove(struct pci_dev *pdev)
{
	struct nvme_cmb_dev *dev = dev_get_drvdata(&pdev->dev);

	dev_info(&pdev->dev, "unbinding NVMe device\n");

	del_gendisk(dev->disk);
	put_disk(dev->disk);
	ida_free(&nvme_cmb_index_ida, dev->index);

	iounmap(dev->buffer_bar);
	nvme_cmb_shutdown(dev);
	nvme_cmb_free_queues(dev, pdev);
	iounmap(dev->bar);
	pci_release_mem_regions(pdev);
	pci_disable_device(pdev);
	kfree(dev);
}

static const struct pci_device_id nvme_cmb_id_table[] = {
	{ PCI_DEVICE_CLASS(PCI_CLASS_STORAGE_EXPRESS, 0xffffff) },
	{ 0, }
};

static struct pci_driver nvme_cmb_driver = {
	.name = "nvme_cmb",
	.id_table = nvme_cmb_id_table,
	.probe = nvme_cmb_probe,
	.remove = nvme_cmb_remove,
};

static int __init nvme_cmb_init(void)
{
	pr_info("loading nvme-cmb module\n");
	return pci_register_driver(&nvme_cmb_driver);
}

static void __exit nvme_cmb_exit(void)
{
	pr_info("unloading nvme-cmb module\n");
	pci_unregister_driver(&nvme_cmb_driver);
}

module_init(nvme_cmb_init);
module_exit(nvme_cmb_exit);
